/**
 *
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

// AMQ Ajax handler
// This class provides the main API for using the Ajax features of AMQ. It
// allows JMS messages to be sent and received from javascript when used
// with the org.apache.activemq.web.MessageListenerServlet.
//
// This version of the file provides an adapter interface for the jquery library
// and a namespace for the Javascript file, private/public variables and
// methods, and other scripting niceties. -- jim cook 2007/08/28

var org = org || {};
org.activemq = org.activemq || {};

org.activemq.AmqAdapter = {
	isTIP: false,
	
/**
 * Initialize the adapter
 */
	init: function(options) {
		// determine if we are running in TIP
		try {
			if (tipdojo) {
				this.isTIP = true;
			}
		}
		catch (e) {
			this.isTIP = false;	
		}
	},

/**
 *  Implement this method to make an AJAX call to the AjaxServlet. An
 *  options object will accompany this class and will contain the properties
 *  that describe the details of the AJAX call. The options object will
 *  have the following properties:
 *
 *  - method:  'get' or 'post'
 *  - data:    query data to accompany the post or get.
 *  - success: A callback function that is invoked upon successful
 *             completion of the AJAX call. The parameter is:
 *             - data: The result of the AJAX call. In the case of XML
 *                     data should resolve to a Document element.
 *  - error:   A callback when some type of error occurs. The callback
 *             function's parameters should be:
 *             - xhr:    The XmlHttpRequest object.
 *             - status: A text string of the status.
 *             - ex:     The exception that caused the error.
 */
	ajax: function(uri, options) {
		// post
		if (options.method == 'post') {
			if (options.data)
			{	uri += "?";
				uri += options.data;
			}
			
			if (this.isTIP) {
				tipdojo.xhr("post", {
					url: uri,
					sync: false,
					handleAs: "xml",										
					load : options.success ? options.success : function() {},
					error: options.error ? function(ex, ioargs) {
							options.error(ioargs.xhr,ioargs.xhr.status, ex);
						} : function() {}				
				});
			}
			else {
				dojo.xhr("post", {
					url: uri,
					sync: false,
					handleAs: "xml",										
					load : options.success ? options.success : function() {},
					error: options.error ? function(ex, ioargs) {
							options.error(ioargs.xhr,ioargs.xhr.status, ex);
						} : function() {}				
				});
			}
		}
		// get
		else {
			if (options.data)
			{	uri += "?";
				uri += options.data;
			}
			
			if (this.isTIP) {
				tipdojo.xhrGet({
					url: uri,
					sync: false,
					handleAs: "xml",				
					load : options.success ? options.success : function() {},
					error: options.error ? function(ex, ioargs) {
							options.error(ioargs.xhr,ioargs.xhr.status, ex);
						} : function() {}
				});
			}
			else {
				dojo.xhrGet({
					url: uri,
					sync: false,
					handleAs: "xml",				
					load : options.success ? options.success : function() {},
					error: options.error ? function(ex, ioargs) {
							options.error(ioargs.xhr,ioargs.xhr.status, ex);
						} : function() {}
				});
			}
		}		
	},

	log: function(message, exception) {
		if (typeof console != 'undefined' && console.log) {
			console.log(message);
		}
	}
};



org.activemq.Amq = function() {
	var connectStatusHandler = null;
	
	// Just a shortcut to eliminate some redundant typing.
	var adapter = org.activemq.AmqAdapter;

	if (typeof adapter == 'undefined') {
		throw 'An org.activemq.AmqAdapter must be declared before the amq.js script file.';
	}

	// Indicates whether init() has been invoked
	var initialized = false;
	
	// Indicates whether terminate() has been invoked
	var terminated = false;

	// Inidicates whether logging is active or not. Not by default.
	var logging = false;
	
	// The URI of the AjaxServlet.
	var uri;

	// The number of seconds that the long-polling socket will stay connected.
	// Best to keep this to a value less than one minute.
	var timeout = 25;
	
	// Poll delay. if set to positive integer, this is the time to wait in ms
	// before sending the next poll after the last completes.
	var pollDelay = 0;

	// 5 second delay if an error occurs during poll. This could be due to
	// server capacity problems or a timeout condition.
	var pollErrorDelay = 5000;

	// Map of handlers that will respond to message receipts. The id used during
	// addListener(id, destination, handler) is used to key the callback
	// handler.  
	var messageHandlers = {};
	
	// Indicates whether an AJAX post call is in progress.
	var batchInProgress = false;

	// A collection of pending messages that accumulate when an AJAX call is in
	// progress. These messages will be delivered as soon as the current call
	// completes. The array contains objects in the format { destination,
	// message, messageType }.
	var messageQueue = [];

	/**
	 * Iterate over the returned XML and for each message in the response, 
	 * invoke the handler with the matching id.
	 */
	var messageHandler = function(data) {
		var response = data.getElementsByTagName("ajax-response");
		if (response !== null && response.length == 1) {
			if (connectStatusHandler) {
				connectStatusHandler(true);
			}
			var responses = response[0].childNodes;    // <response>
			for (var i = 0; i < responses.length; i++) {
				var responseElement = responses[i];

				// only process nodes of type element.....
				if (responseElement.nodeType != 1) {
					continue;
				}

				var id = responseElement.getAttribute('id');				
				var handler = messageHandlers[id];				
				if (handler) {
					// Loop thru and handle each <message>
					for (var j = 0; j < responseElement.childNodes.length; j++) {
						var msg = responseElement.childNodes[j];
						handler(msg);					
					}
				}
				else if (logging) {
					adapter.log('No handler found to match message with id = ' + id);
				}
			}
		}
	};

	var errorHandler = function(xhr, status, ex) {		
		if (connectStatusHandler) {
			connectStatusHandler(false);
		}
		if (logging) {
			adapter.log('Error occurred in ajax call. HTTP result: ' + xhr.status + ', status: ' + status);
		}
	};

	var pollErrorHandler = function(xhr, status, ex) {	
		if (connectStatusHandler) {
			connectStatusHandler(false);
		}
		
		if (status === 'error' && xhr.status === 0) {
			if (logging) {
				adapter.log('Server connection dropped.');
			}			
		}
		else if (logging) {
				adapter.log('Error occurred in poll. HTTP result: ' + xhr.status + ', status: ' + status);
		}
		
		setTimeout(function() { sendPoll(); }, pollErrorDelay);
	};

	var pollHandler = function(data) {
		try {		
			messageHandler(data);
		} 
		catch(e) {
			if (logging) {
				adapter.log('Exception in the poll handler: ' + data, e);
			}
			//throw(e);
		} 
		finally {
			setTimeout(function() { sendPoll(); }, pollDelay);
		}		
	};

	var sendPoll = function() {
		if (!terminated) {
			// Workaround IE6 bug where it caches the response
			// Generate a unique query string with date and random
			var now = new Date();		
			var data = 'timeout=' + (timeout * 1000) + '&d=' + now.getTime() + '&r=' + Math.random();
				 		
			var options = { method: 'get',
				data: data,
				success: pollHandler,
				error: pollErrorHandler};
			adapter.ajax(uri, options);	
		}
		else { 
			terminated = false;
		}
	};

	startBatch = function() {
		batchInProgress = true;
	};

	endBatch = function() {
		if (messageQueue.length > 0) {
			var body = buildParams(messageQueue);
			messageQueue.length = 0;
			startBatch();
			adapter.ajax(uri, {
				method: 'post',
				data: body,
				success: endBatch, 
				error: errorHandler});
		} else {
			batchInProgress = false;
		}
	};
		
	var sendJmsMessage = function(destination, message, type) {
		// Add message to outbound queue
		if (batchInProgress) {
			messageQueue[messageQueue.length] = {
				destination: destination,
				message: message,
				messageType: type
			};
		} else {
			startBatch();
			adapter.ajax(uri, { method: 'post',
				data: 'destination=' + destination + '&message=' + message + '&type=' + type,
				error: function(xhr, status, ex) {
					batchInProgress = false;
					errorHandler(xhr, status, ex);								
				},
				success: endBatch});
		}
	};

	var buildParams = function(msgs) {
		var s = [];
		for (var i = 0, c = msgs.length; i < c; i++) {
			if (i !== 0) {
				s[s.length] = '&';
			}
			
			s[s.length] = ((i === 0) ? 'destination' : 'd' + i);
			s[s.length] = '=';
			s[s.length] = msgs[i].destination;
			s[s.length] = ((i === 0) ? '&message' : '&m' + i);
			s[s.length] = '=';
			s[s.length] = msgs[i].message;
			s[s.length] = ((i === 0) ? '&type' : '&t' + i);
			s[s.length] = '=';
			s[s.length] = msgs[i].messageType;
		}
		return s.join('');
	};

	return {
		isInitialized: function() {
			return initialized;
		},

		init: function(options) {
			initialized = true;
			
			// set up init options
			connectStatusHandler = options.connectStatusHandler;
			uri = options.uri || '/amq';
			timeout = typeof options.timeout == 'number' ? options.timeout : 25;
			pollDelay = typeof options.pollDelay == 'number' ? options.pollDelay : 0;
			pollErrorDelay = typeof options.pollErrorDelay == 'number' ? options.pollErrorDelay : 5000;
			logging = options.logging;
			
			// initialize ajax adapter
			adapter.init(options);
			
			// initiate polling
			sendPoll();					
		},

		terminate: function() {
			terminated = true;
			initialized = false;			
		},
		
		setConnectStatusHandler: function(handler) {
			connectStatusHandler = handler;
		},

		// Send a JMS message to a destination (eg topic://MY.TOPIC).  Message
		// should be xml or encoded xml content.
		sendMessage : function(destination, message) {
			sendJmsMessage(destination, message, 'send');
		},

		// Listen on a channel or topic.
		// handler must be a function taking a message argument
		addListener : function(id, destination, handler) {
			messageHandlers[id] = handler;
			sendJmsMessage(destination, id, 'listen');		
		},

		// Remove Listener from channel or topic.
		removeListener : function(id, destination) {
			delete messageHandlers[id];
			sendJmsMessage(destination, id, 'unlisten');
		}
	};
}();














function ibm_tivoli_navmodel_EventRouter() {}

// data members
ibm_tivoli_navmodel_EventRouter.initialized = false;
ibm_tivoli_navmodel_EventRouter.isTIP = false;
ibm_tivoli_navmodel_EventRouter.eventDestination = null;
ibm_tivoli_navmodel_EventRouter.timeout = 15;
ibm_tivoli_navmodel_EventRouter.pollDelay = 0;
ibm_tivoli_navmodel_EventRouter.pollErrorDelay = 5000;
ibm_tivoli_navmodel_EventRouter.contextRoot = "/ibm/tivoli/amq";
ibm_tivoli_navmodel_EventRouter.registry = {};


/*
 * Returns true if already initialized
 */
ibm_tivoli_navmodel_EventRouter.isInitialized = function() {
	return this.initialized;
};


/*
 * Initializes connection with event feed
 */
ibm_tivoli_navmodel_EventRouter.init = function(args) {
	if (!this.initialized) {
		this.initialized = true;
		
		this.eventDestination = args.eventDestination;
		
		if (args.timeout !== undefined) {
			this.timeout = args.timeout;
		}
		
		if (args.pollDelay !== undefined) {
			this.pollDelay = args.pollDelay;
		}
		
		if (args.pollErrorDelay !== undefined) {
			this.pollErrorDelay = args.pollErrorDelay;
		}
		
		if (args.contextRoot !== undefined) {
			this.contextRoot = args.contextRoot + "/amq";
		}
		
		// determine if we are running in TIP
		try {
			if (tipdojo) {
				this.isTIP = true;			
			}
		}
		catch (e) {
			this.isTIP = false;		
		}
	}
	
	// initialize amq
	if (!org.activemq.Amq.isInitialized()) {							
		var initArgs = {
			uri: this.contextRoot, 
			timeout: this.timeout,
			pollDelay: this.pollDelay,
			pollErrorDelay: this.pollErrorDelay,
			connectStatusHandler: ibm_tivoli_navmodel_EventRouter.connectionStatus,
			logging: true
		};
			
		this.info("***** event_router - initializing event feed");
		org.activemq.Amq.init(initArgs);	
		org.activemq.Amq.addListener("ibm_tivoli_navmodel_EventRouter", this.eventDestination, ibm_tivoli_navmodel_EventRouter.routeEvent);	
	}
};


/*
 * Terminates connection with event feed
 */
ibm_tivoli_navmodel_EventRouter.terminate = function() {
	if (org.activemq.Amq.isInitialized()) {							
		this.info("***** event_router - terminating event feed");
		org.activemq.Amq.removeListener("ibm_tivoli_navmodel_EventRouter", this.eventDestination);
		org.activemq.Amq.terminate();				
	}
};


/*
 * Adds listener for the requested eventId
 */
ibm_tivoli_navmodel_EventRouter.addListener = function(listener, eventId, eventCallback) {
	// add entry if it doesn't already exist
	if (!this.registry[eventId]) {		
		this.registry[eventId] = {listeners: []};
	}
		
	// append listener to the entry
	var listenersLength = this.registry[eventId].listeners.length;
	this.info("***** event_router - adding listener for eventId=" + eventId + " at index=" + listenersLength);		
	this.registry[eventId].listeners[listenersLength] = {listener: listener, eventCallback: eventCallback};	
	
	// validate listeners
	this.validate();
};


/*
 * Removes listener for the requested eventId
 */
ibm_tivoli_navmodel_EventRouter.removeListener = function(listener, eventId) {
	var listenersLength = 0;
	
	if (this.registry[eventId]) {
		listenersLength = this.registry[eventId].listeners.length;
		for (var listenerIndex = 0;  listenerIndex < listenersLength;  ++listenerIndex) {
			// remove listener from entry
			if (this.registry[eventId].listeners[listenerIndex].listener === listener) {
				this.info("***** event_router - remove listener for eventId=" + eventId + " at index=" + listenerIndex);
				this.registry[eventId].listeners.splice(listenerIndex, 1);
					
				// remove entry if no more listeners
				listenersLength = this.registry[eventId].listeners.length;
				if (listenersLength === 0) {
					delete this.registry[eventId];
					
					// no more listeners so shut down event feed
					if (!this.haveListeners()) {
						this.terminate();
					}					
				}					
				break;
			}
		}
	}	
		
	return listenersLength;		
};


/*
 * Returns registered listener count
 */
ibm_tivoli_navmodel_EventRouter.getListenerCount = function() {
	var listeners = 0;
	
	for (var eventId in this.registry) {
		if (this.registry.hasOwnProperty(eventId)) {
			var listenerLength = this.registry[eventId].listeners.length;				
			listeners += listenerLength;		
		}
	}
	
	return listeners;
};


/*
 * Returns true if we have at least one listener
 */
ibm_tivoli_navmodel_EventRouter.haveListeners = function() {
	for (var eventId in this.registry) {
		if (this.registry.hasOwnProperty(eventId)) {
			var listenerLength = this.registry[eventId].listeners.length;				
			if (listenerLength > 0) {
				return true;
			}		
		}
	}
	
	return false;
};


/*
 * Validate registered liseners; remove any that can no longer be pinged
 */
ibm_tivoli_navmodel_EventRouter.validate = function() {
	for (var eventId in this.registry) {
		if (this.registry.hasOwnProperty(eventId)) {
			var listenersLength = this.registry[eventId].listeners.length;				
			for (var listenerIndex = listenersLength - 1;  listenerIndex >= 0;  --listenerIndex) {
				var listener = this.registry[eventId].listeners[listenerIndex].listener;
				try {
					listener.ping();								
				}
				catch (e) {	
					this.warn("***** event_router - validate is removing invalid listener ");
					this.removeListener(listener, eventId);
					//invalid[invalid.length] = listener;
				}				
			}
		}
	}
};


/*
 * Routes events from event feed to the listeners;
 * it is being invoked without a "this" so make sure
 * to use the global references to variables
 */
ibm_tivoli_navmodel_EventRouter.routeEvent = function(jmsMsg) {
	if (jmsMsg.data.length > 0) {			
		var event = null;
		if (ibm_tivoli_navmodel_EventRouter.isTIP) {
			event = tipdojo.fromJson(jmsMsg.data);
		}
		else {
			event = dojo.fromJson(jmsMsg.data);
		}
			
		var eventId = event.id;
		var registry = ibm_tivoli_navmodel_EventRouter.registry;	
			
		if (registry[eventId]) {
			var listenersLength = registry[eventId].listeners.length;						
			for (var listenerIndex = listenersLength - 1;  listenerIndex >= 0;  --listenerIndex) {
				try {
					var eventCallback = registry[eventId].listeners[listenerIndex].eventCallback;				
					eventCallback(event);									
				}
				catch (e) {	
					// see if listener is still pingable				
					var listener = registry[eventId].listeners[listenerIndex].listener;
					try {
						listener.ping();
						ibm_tivoli_navmodel_EventRouter.error("***** event_router - Exception when routing eventId=" + eventId + " exception=" + e);					
					}
					catch (e2) {
						// listener is dead we must remove it
						ibm_tivoli_navmodel_EventRouter.warn("***** event_router - routeEvent is removing invalid listener");
						ibm_tivoli_navmodel_EventRouter.removeListener(listener, eventId);
					}
				}				
			}
		}
	}
};


/*
 * Connection status handler;  called from event feed;
 * it is being invoked without a "this" so make sure to
 * use the global references to variables
 */
ibm_tivoli_navmodel_EventRouter.connectionStatus = function(status) {
	//ibm_tivoli_navmodel_EventRouter.info("***** event_router - connection status=" + status);
};


/*
 * Log info message
 */
ibm_tivoli_navmodel_EventRouter.info = function(msg) {
	console.info(msg);
	//alert(msg);
};


/*
 * Log warning message
 */		
ibm_tivoli_navmodel_EventRouter.warn = function(msg) {
	console.warn(msg);
	//alert(msg);
};


/*
 * Log error message
 */	
ibm_tivoli_navmodel_EventRouter.error = function(msg) {
	console.error(msg);
	//alert(msg);
};
	